from threading import Lock
from functools import wraps
from logging import getLogger
import time

from opentsdb.exceptions import TagsError

logger = getLogger('opentsdb-py')


def send_metric(func):

    def validate_tags(tag_values_length, tag_names_length, optional_tags):
        if optional_tags is False and tag_values_length != tag_names_length:
            raise TagsError("Tags is incorrect, expected %d != %d" % (tag_names_length, tag_values_length))

    @wraps(func)
    def wrapper(self: Metric, *args, **kwargs):
        with self._lock:
            try:
                result = func(self, *args, **kwargs)
                tags = self.pop_tags()
                validate_tags(len(tags), self.tag_names_length, self.optional_tags)
                self.client.send(self.name, self._value.get(), **tags)
                return result
            except ValueError:
                pass

    return wrapper


class Metric:

    def __init__(self, name: str, tag_names=(), client=None, optional_tags=False):
        self.name = name

        self.tag_names = tag_names
        self.tag_names_length = len(tag_names)
        self.optional_tags = optional_tags

        self.client = client

        self._value = None
        self._tags = None
        self._lock = Lock()

    @property
    def value(self):
        return self._value.get()

    def tags(self, *tag_values):
        if len(tag_values) > self.tag_names_length:
            raise TagsError("Too many tags is set, expected %d" % self.tag_names_length)

        self._tags = dict(zip(self.tag_names, tag_values))

        return self

    def pop_tags(self) -> dict:
        tags = self._tags or {}
        self._tags = None
        return tags


class _MetricValue:
    def __init__(self):
        self._value = 0.0
        self._lock = Lock()

    def inc(self, amount):
        with self._lock:
            self._value += amount

    def set(self, value):
        with self._lock:
            self._value = value

    def get(self):
        with self._lock:
            return self._value


class Counter(Metric):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = _MetricValue()
        logger.info("Metric registered [type: Counter]: %s", self.name)

    @send_metric
    def inc(self, amount=1):
        if amount < 0:
            raise ValueError('Counters can only be incremented by non-negative amounts.')
        self._value.inc(amount)

    def __str__(self):
        return "%s %s" % (self.name, self._value.get())


class Gauge(Metric):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = _MetricValue()
        logger.info("Metric registered [type: Gauge]: %s", self.name)

    @send_metric
    def inc(self, amount=1):
        self._value.inc(amount)

    @send_metric
    def dec(self, amount=1):
        self._value.inc(-amount)

    @send_metric
    def set(self, value: float):
        self._value.set(float(value))

    @send_metric
    def timeit(self, timer=time.perf_counter):
        return _GaugeTimer(self, timer)


class _GaugeTimer:
    def __init__(self, gauge, timer):
        self._gauge = gauge
        self._timer = timer

    def __enter__(self):
        self._start = self._timer()

    def __exit__(self, *_):
        self._gauge.set(max(self._timer() - self._start, 0))

    def __call__(self, func):
        def wrapped(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return wrapped



import threading
import logging
import socket
import string
import queue
import time
from os import environ

from opentsdb.exceptions import MetricDuplicated, ValidationError
from opentsdb.tsdb_connect import TSDBConnect
from opentsdb.push_thread import PushThread
from opentsdb.metrics import Metric

logger = logging.getLogger('opentsdb-py')


class TSDBClient:
    TSDB_HOST = environ.get('OPEN_TSDB_HOST', '127.0.0.1')
    TSDB_PORT = environ.get('OPEN_TSDB_PORT', 4242)

    MAX_METRICS_QUEUE_SIZE = 10000
    SEND_METRICS_PER_SECOND_LIMIT = 1000
    VALID_METRICS_CHARS = set(string.ascii_letters + string.digits + '-_./')

    def __init__(self, host: str=TSDB_HOST, port: int=TSDB_PORT, check_tsdb_alive: bool=False, run_at_once: bool=True,
                 static_tags: dict=None, host_tag: bool=True, test_mode: bool=False, raise_duplicate=True,
                 max_queue_size: int=MAX_METRICS_QUEUE_SIZE, send_metrics_limit: int=SEND_METRICS_PER_SECOND_LIMIT):
        self.host_tag = host_tag
        self.check_tsdb_alive = check_tsdb_alive
        self.static_tags = static_tags or {}
        self.raise_duplicate = raise_duplicate
        self.send_metrics_limit = send_metrics_limit
        self.test_mode = test_mode

        self._tsdb_connect = None
        self._close_client = threading.Event()

        self._last_metric = None
        self._metrics_queue = queue.Queue(maxsize=max_queue_size)

        self._metric_send_thread = None

        if run_at_once is True:
            self.init_client(host, port)

    def init_client(self, host, port: int=TSDB_PORT):
        self._tsdb_connect = TSDBConnect(host, port, self.check_tsdb_alive)

        self._metric_send_thread = PushThread(
            self._tsdb_connect, self._metrics_queue, self._close_client, self.send_metrics_limit, self.test_mode)
        self._metric_send_thread.daemon = True
        self._metric_send_thread.start()

    def __getattr__(self, item):
        return getattr(self, item)

    def __setattr__(self, key, value):
        if isinstance(value, Metric):
            if value.client is None:
                value.client = self

        super(TSDBClient, self).__setattr__(key, value)

    def is_connected(self) -> bool:
        return self._tsdb_connect.is_alive()

    def close(self):
        self._close_client.set()

    def wait(self):
        while self._metric_send_thread.is_alive():
            time.sleep(0.05)

    def stop(self):
        self._tsdb_connect.stopped.set()

    def queue_size(self) -> int:
        return self._metrics_queue.qsize()

    def send(self, name: str, value, **tags) -> str:
        tags.update(self.static_tags)
        if self.host_tag is True and 'host' not in tags:
            tags['host'] = socket.gethostname()

        self._validate_metric(name, value, tags)

        timestamp = int(tags.pop('timestamp', time.time()))
        tags_string = ' '.join(['%s=%s' % (key, value) for key, value in tags.items()])
        metric_str = "put %s %d %s %s\n" % (name, timestamp, value, tags_string)

        self._check_duplicate(metric_str)
        if not self._close_client.is_set():
            self._push_metric_to_queue(metric_str)

        return metric_str

    def _validate_metric(self, name, value, tags):
        try:
            assert all(char in self.VALID_METRICS_CHARS for char in name), \
                "Metric name contain incorrect chars '%s'" % name
            assert isinstance(value, (str, int, float)), "Incorrect metric value type '%s'" % type(value)
            assert tags != {}, "Need at least one tag"
        except AssertionError as error:
            raise ValidationError("Metric not valid: %s" % str(error))

    def _check_duplicate(self, metric_str):
        try:
            assert metric_str != self._last_metric
            self._last_metric = metric_str
        except AssertionError:
            if self.raise_duplicate:
                raise MetricDuplicated("Duplicate metric: %s" % metric_str)
            logger.warning("Duplicate metric: %s" % metric_str)

    def _push_metric_to_queue(self, metric_str):
        try:
            self._metrics_queue.put(metric_str, False)
        except queue.Full:
            logger.warning("Drop oldest metric because Queue is full.")
            self._metrics_queue.get()
            self._metrics_queue.put(metric_str, False)

    def __del__(self):
        self.close()



from logging import getLogger
from queue import Empty
import threading
import random
import time

from opentsdb.exceptions import TSDBConnectionError

logger = getLogger('opentsdb-py')


class PushThread(threading.Thread):

    def __init__(self, tsdb_connect, metrics_queue, close_client, send_metrics_limit, test_mode):
        super().__init__()
        self.tsdb_connect = tsdb_connect
        self.metrics_queue = metrics_queue
        self.close_client_flag = close_client
        self.send_metrics_limit = send_metrics_limit
        self.test_mode = test_mode

        self.retry_send_metric = None

    def run(self):
        while not self._is_done():
            start_time = time.time()

            try:
                metric = self._next_metric()
            except Empty:
                continue

            logger.debug("Send metric: %s", metric)
            if not self.test_mode:
                try:
                    self.tsdb_connect.sendall(metric.encode('utf-8'))
                except TSDBConnectionError:
                    self.retry_send_metric = metric
                    time.sleep(0.5)
                except Exception as error:
                    logger.exception("Push metric failed: %s", error)
                    self.retry_send_metric = metric
                    time.sleep(1)

            self.__metrics_limit_timeout(start_time)

        self.tsdb_connect.disconnect()

    def _is_done(self):
        return self.tsdb_connect.stopped.is_set() or (self.close_client_flag.is_set() and self.metrics_queue.empty())

    def _next_metric(self, wait_timeout=3):
        if self.retry_send_metric:
            metric = self.retry_send_metric
            self.retry_send_metric = None
        else:
            metric = self.metrics_queue.get(block=True, timeout=wait_timeout)

        return metric

    def __metrics_limit_timeout(self, start_time):
        duration = time.time() - start_time
        if self.send_metrics_limit > 0:
            wait_time = (2.0 * random.random()) / self.send_metrics_limit
            if wait_time > duration:
                logger.debug("Wait for %s", wait_time - duration)
                time.sleep(wait_time - duration)



from setuptools import setup, find_packages

from opentsdb import __version__


setup(
    name='opentsdb-py',
    version=__version__,
    author='Sergey Suglobov',
    author_email='s.suglobov@gmail.com',
    packages=find_packages(),
    keywords="opentsdb, tsdb, metrics",
    url='https://github.com/scarchik/opentsdb-py',
    download_url='https://github.com/scarchik/opentsdb-py/archive/v{VERSION}.zip'.format(VERSION=__version__),
    description='Python3 client for OpenTSDB',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only'
    ],
    install_requires=[
        "setuptools",
    ]
)



import threading
import logging
import socket
import time

from opentsdb.exceptions import TSDBConnectionError

logger = logging.getLogger('opentsdb-py')


class TSDBConnect:

    def __init__(self, host: str, port: int, check_tsdb_alive: bool=False):
        self.tsdb_host = host
        self.tsdb_port = int(port)

        if check_tsdb_alive:
            self.is_alive(raise_error=True)

        self._connect = None
        self.stopped = threading.Event()

    def is_alive(self, timeout=3, raise_error=False) -> bool:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(timeout)
            sock.connect((self.tsdb_host, self.tsdb_port))
            sock.close()
        except (ConnectionRefusedError, socket.timeout):
            if raise_error is True:
                raise
            return False
        else:
            return True

    @property
    def connect(self) -> socket.socket:
        if not self._connect or getattr(self._connect, '_closed', False) is True:
            logger.debug("Connect to OpenTSDB: %s:%s", self.tsdb_host, self.tsdb_port)
            self.stopped.clear()
            self._make_connection()

        return self._connect

    def _make_connection(self, timeout=2):
        attempt = 0
        while not self.stopped.is_set():
            try:
                self._connect = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self._connect.settimeout(timeout)
                self._connect.connect((self.tsdb_host, self.tsdb_port))
                return
            except (ConnectionRefusedError, socket.timeout):
                time.sleep(min(15, 2 ** attempt))
                attempt += 1

    def disconnect(self):
        logger.debug("Disconnecting from %s:%s", self.tsdb_host, self.tsdb_port)
        self.stopped.set()
        if self._connect:
            self._connect.close()
        self._connect = None

    def sendall(self, line: bytes):
        try:
            self.connect.sendall(line)
        except (BrokenPipeError, IOError, OSError, socket.error) as error:
            logger.exception("Close connection to handle exception: %s", error)
            self._connect.close()
            raise TSDBConnectionError(str(error))





class TSDBClientException(Exception):
    pass


class TSDBConnectionError(TSDBClientException):
    pass


class TagsError(TSDBClientException):
    pass


class MetricDuplicated(TSDBClientException):
    pass


class ValidationError(TSDBClientException):
    pass
